@@ -185,9 +185,12 @@ class PythonToPythonJS(NodeVisitor):
         self._global_typed_lists = dict()  ## global name : set  (if len(set)==1 then we know it is a typed list)
         self._global_typed_dicts = dict()
         self._global_typed_tuples = dict()
+        self._global_functions = set()
 
         self._custom_operators = {}
-
+        self._injector = []  ## advanced meta-programming hacks
+        self._in_class = None
+        self._with_fastdef = False
         self.setup_builtins()
 
     def preprocess_custom_operators(self, data):
@@ -215,7 +218,7 @@ class PythonToPythonJS(NodeVisitor):
         self._classes['dict'] = set(['__getitem__', '__setitem__'])
         self._classes['list'] = set(['__getitem__', '__setitem__'])
         self._classes['tuple'] = set(['__getitem__', '__setitem__'])
-        self._builtins = set(['dict', 'list', 'tuple'])
+        self._builtin_classes = set(['dict', 'list', 'tuple'])
 
     def get_typedef(self, instance=None, class_name=None):
         assert instance or class_name
@@ -405,6 +408,7 @@ class PythonToPythonJS(NodeVisitor):
     def visit_ClassDef(self, node):
         name = node.name
         log('ClassDef: %s'%name)
+        self._in_class = name
         self._classes[ name ] = list()  ## method names
         self._class_parents[ name ] = set()
         self._class_attributes[ name ] = set()
@@ -413,9 +417,21 @@ class PythonToPythonJS(NodeVisitor):
         self._decorator_class_props[ name ] = self._decorator_properties
         self._instances[ 'self' ] = name
 
-        #for dec in node.decorator_list:  ## TODO class decorators
-        #    if isinstance(dec, Name) and dec.id == 'inline':  ## @inline class decorator is DEPRECATED
-        #        self._catch_attributes = set()
+        class_decorators = []
+        self._injector = []
+        for decorator in node.decorator_list:  ## class decorators
+            if isinstance(decorator, Attribute) and isinstance(decorator.value, Name) and decorator.value.id == 'pythonjs':
+                if decorator.attr == 'property_callbacks':
+                    self._injector.append('set')
+                elif decorator.attr == 'init_callbacks':
+                    self._injector.append('init')
+                else:
+                    raise SyntaxError( 'unsupported pythonjs class decorator' )
+
+            else:
+                #raise SyntaxError( 'unsupported class decorator' )
+                class_decorators.append( decorator )
+
         ## always catch attributes ##
         self._catch_attributes = set()
         self._instance_attributes[ name ] = self._catch_attributes
@@ -438,8 +454,8 @@ class PythonToPythonJS(NodeVisitor):
             if isinstance(item, FunctionDef):
                 log('  method: %s'%item.name)
 
-                if item.name == '__contains__':  ## this is required because we added Object.prototype.__contains__
-                    item.name = item.name.upper()
+                #if item.name == '__contains__':  ## this is required because we added Object.prototype.__contains__ - DEPRECATED!
+                #    item.name = item.name.upper()
 
                 self._classes[ name ].append( item.name )
                 item_name = item.name
@@ -468,13 +484,24 @@ class PythonToPythonJS(NodeVisitor):
 
         for prop_name in self._decorator_properties:
             getter = self._decorator_properties[prop_name]['get']
-            writer.write('window["__%s_properties"]["%s"] = %s' %(name, prop_name, getter))
+            writer.write('window["__%s_properties"]["%s"] = JSObject()' %(name, prop_name))
+            writer.write('window["__%s_properties"]["%s"]["get"] = %s' %(name, prop_name, getter))
+            if self._decorator_properties[prop_name]['set']:
+                setter = self._decorator_properties[prop_name]['set']
+                writer.write('window["__%s_properties"]["%s"]["set"] = %s' %(name, prop_name, setter))
 
         self._catch_attributes = None
         self._decorator_properties = None
         self._instances.pop('self')
+        self._in_class = False
 
         writer.write('%s = create_class("%s", window["__%s_parents"], window["__%s_attrs"], window["__%s_properties"])' % (name, name, name, name, name))
+        if 'init' in self._injector:
+            writer.write('%s.init_callbacks = JSArray()' %name)
+        self._injector = []
+
+        for dec in class_decorators:
+            writer.write('%s = get_attribute(%s,"__call__")( [%s], JSObject() )' % (name, self.visit(dec), name))
 
     def visit_And(self, node):
         return ' and '
@@ -527,6 +554,8 @@ class PythonToPythonJS(NodeVisitor):
                 return 'true'
             elif node.id == 'False':
                 return 'false'
+            elif node.id == 'None':
+                return 'undefined'
 
         return node.id
 
@@ -540,7 +569,11 @@ class PythonToPythonJS(NodeVisitor):
             elif isinstance(node.value, Name) and node.value.id == 'self' and 'self' in self._instances:
                 self._return_type = self._instances['self']
 
-            writer.write('return %s' % self.visit(node.value))
+            if self._cached_property:
+                writer.write('self["__dict__"]["%s"] = %s' %(self._cached_property, self.visit(node.value)) )
+                writer.write('return self["__dict__"]["%s"]' %self._cached_property)
+            else:
+                writer.write('return %s' % self.visit(node.value))
 
         else:
             writer.write('return')  ## empty return
@@ -655,7 +688,14 @@ class PythonToPythonJS(NodeVisitor):
         return ' is not '
 
     def visit_UnaryOp(self, node):
-        return self.visit(node.op) + self.visit(node.operand)
+        op = self.visit(node.op)
+        if op is None: raise RuntimeError( node.op )
+        operand = self.visit(node.operand)
+        if operand is None: raise RuntimeError( node.operand )
+        return op + operand
+
+    def visit_USub(self, node):
+        return '-'
 
     def visit_Attribute(self, node):
         node_value = self.visit(node.value)
@@ -936,12 +976,17 @@ class PythonToPythonJS(NodeVisitor):
                             raise SyntaxError('global lists can only contain one type: instance "%s" is unknown' %node.args[0].id)
 
             call_has_args_only = len(node.args) and not (len(node.keywords) or node.starargs or node.kwargs)
+            call_has_args_kwargs_only = len(node.args) and len(node.keywords) and not (node.starargs or node.kwargs)
             call_has_args = len(node.args) or len(node.keywords) or node.starargs or node.kwargs
             name = self.visit(node.func)
 
             if call_has_args_only:  ## lambda only supports simple args for now.
                 args = ', '.join(map(self.visit, node.args))
 
+            elif call_has_args_kwargs_only:
+                args = ', '.join(map(self.visit, node.args))
+                kwargs = ', '.join(map(lambda x: '%s:%s' % (x.arg, self.visit(x.value)), node.keywords))
+
             elif call_has_args:
                 args = ', '.join(map(self.visit, node.args))
                 kwargs = ', '.join(map(lambda x: '%s=%s' % (x.arg, self.visit(x.value)), node.keywords))
@@ -952,7 +997,6 @@ class PythonToPythonJS(NodeVisitor):
                 self.identifier += 1
 
                 if name in ('list', 'tuple'):
-                    #writer.append( '%s = JS("%s.__dict__.js_object")' % (args_name, args))
                     if args:
                         writer.append( '%s = %s.__dict__.js_object' % (args_name, args))  ## test this
                     else:
@@ -970,7 +1014,17 @@ class PythonToPythonJS(NodeVisitor):
                     writer.append(code)
 
             if call_has_args_only:
-                return 'get_attribute(%s, "__call__")([%s], JSObject())' % (name, args)
+                if name in self._global_functions:
+                    return '%s( [%s], __NULL_OBJECT__ )' %(name,args)
+                else:
+                    return 'get_attribute(%s, "__call__")([%s], __NULL_OBJECT__)' % (name, args)
+
+            elif call_has_args_kwargs_only:
+                if name in self._global_functions:
+                    return '%s( [%s], {%s} )' %(name,args, kwargs)
+                else:
+                    return 'get_attribute(%s, "__call__")([%s], {%s} )' % (name, args, kwargs)
+
 
             elif call_has_args:
                 if name == 'dict':
@@ -983,19 +1037,28 @@ class PythonToPythonJS(NodeVisitor):
                 else:
                     return 'get_attribute(%s, "__call__")(%s, %s)' % (name, args_name, kwargs_name)
 
-            elif name in self._classes or name in self._builtins:
-                return 'get_attribute(%s, "__call__")( JSArray(), JSObject() )' %name
+            elif name in self._classes:
+                return 'get_attribute(%s, "__call__")( )' %name
+
+            elif name in self._builtin_classes:
+                return 'get_attribute(%s, "__call__")( )' %name
+
+            elif name in self._global_functions:
+                #return 'get_attribute(%s, "__call__")( JSArray(), JSObject() )' %name  ## SLOW ##
+                return '%s( )' %name  ## this is much FASTER ##
 
             else:
-                ## this a slightly dangerous optimization,
-                ## because if the user is trying to create an instance of some class
+                ## if the user is trying to create an instance of some class
                 ## and that class is define in an external binding,
                 ## and they forgot to put "from mylibrary import *" in their script (an easy mistake to make)
                 ## then this fails to call __call__ to initalize the instance,
-                ## and throws a confusing error:
+                ## or a factory function was used that was passed the class to make,
+                ## it will throw this confusing error:
                 ## Uncaught TypeError: Property 'SomeClass' of object [object Object] is not a function 
                 ## TODO - remove this optimization, or provide the user with a better error message.
-                return '%s( JSArray(), JSObject() )' %name
+
+                ## So to be safe we still wrap with get_attribute and "__call__"
+                return 'get_attribute(%s, "__call__")( )' %name
 
     def visit_Lambda(self, node):
         args = [self.visit(a) for a in node.args.args]
@@ -1008,16 +1071,33 @@ class PythonToPythonJS(NodeVisitor):
         property_decorator = None
         decorators = []
         with_js_decorators = []
+        setter = False
+        return_type = None
+        fastdef = False
+        javascript = False
+        self._cached_property = None
+
+        if writer.is_at_global_level():
+            self._global_functions.add( node.name )
+
         for decorator in reversed(node.decorator_list):
             log('@decorator: %s' %decorator)
             if self._with_js:  ## decorators are special in with-js mode
                 with_js_decorators.append( self.visit( decorator ) )
 
-            elif isinstance(decorator, Name) and decorator.id == 'property':
+            elif isinstance(decorator, Name) and decorator.id == 'fastdef':
+                fastdef = True
+
+            elif isinstance(decorator, Name) and decorator.id == 'javascript':
+                javascript = True
+
+            elif isinstance(decorator, Name) and decorator.id in ('property', 'cached_property'):
                 property_decorator = decorator
                 n = node.name + '__getprop__'
                 self._decorator_properties[ node.original_name ] = dict( get=n, set=None )
                 node.name = n
+                if decorator.id == 'cached_property':
+                    self._cached_property = node.original_name
 
             elif isinstance(decorator, Attribute) and isinstance(decorator.value, Name) and decorator.value.id in self._decorator_properties:
                 if decorator.attr == 'setter':
@@ -1026,6 +1106,9 @@ class PythonToPythonJS(NodeVisitor):
                     n = node.name + '__setprop__'
                     self._decorator_properties[ decorator.value.id ]['set'] = n
                     node.name = n
+                    setter = True
+                    prop_name = node.original_name
+
                 elif decorator.attr == 'deleter':
                     raise NotImplementedError
                 else:
@@ -1039,14 +1122,17 @@ class PythonToPythonJS(NodeVisitor):
                     raise RuntimeError( op, self._custom_operators )
                 self._custom_operators[ op ] = node.name
 
+            elif isinstance(decorator, Call) and decorator.func.id == 'returns':
+                assert len(decorator.args) == 1
+                assert isinstance( decorator.args[0], Name)
+                return_type = decorator.args[0].id
+
             else:
                 decorators.append( decorator )
 
         log('function: %s'%node.name)
-        if self._with_js:
-            if node.args.defaults:
-                raise SyntaxError( 'pure javascript functions can not take keyword arguments')
-            elif node.args.vararg:
+        if self._with_js or javascript:
+            if node.args.vararg:
                 raise SyntaxError( 'pure javascript functions can not take variable arguments (*args)' )
             elif node.args.kwarg:
                 raise SyntaxError( 'pure javascript functions can not take variable keyword arguments (**kwargs)' )
@@ -1062,25 +1148,53 @@ class PythonToPythonJS(NodeVisitor):
         ## this is kept here as an option to be sure we are compatible with the
         ## old-style code in runtime/pythonpythonjs.py and runtime/builtins.py
         if not GLOBAL_VARIABLE_SCOPE:
-            local_vars = set()
-            global_vars = set()
-            for n in node.body:
-                if isinstance(n, Assign) and isinstance(n.targets[0], Name):  ## assignment to local
-                    local_vars.add( n.targets[0].id )
-                elif isinstance(n, Global):
-                    global_vars.update( n.names )
-                elif isinstance(n, With) and isinstance( n.context_expr, Name ) and n.context_expr.id == 'javascript':
-                    for c in n.body:
-                        if isinstance(c, Assign) and isinstance(c.targets[0], Name):  ## assignment to local
-                            local_vars.add( c.targets[0].id )
+            def retrieve_vars(body):
+                local_vars = set()
+                global_vars = set()
+                for n in body:
+                    if isinstance(n, Assign) and isinstance(n.targets[0], Name):  ## assignment to local
+                        local_vars.add( n.targets[0].id )
+                    elif isinstance(n, Global):
+                        global_vars.update( n.names )
+                    elif isinstance(n, With) and isinstance( n.context_expr, Name ) and n.context_expr.id == 'javascript':
+                        for c in n.body:
+                            if isinstance(c, Assign) and isinstance(c.targets[0], Name):  ## assignment to local
+                                local_vars.add( c.targets[0].id )
+                    elif hasattr(n, 'body') and not isinstance(n, FunctionDef):
+                        # do a recursive search inside new block except function def
+                        l, g = retrieve_vars(n.body)
+                        local_vars.update(l)
+                        global_vars.update(g)
+                        if hasattr(n, 'orelse'):
+                            l, g = retrieve_vars(n.orelse)
+                            local_vars.update(l)
+                            global_vars.update(g) 
+                return local_vars, global_vars
 
+            local_vars, global_vars = retrieve_vars(node.body)
             if local_vars-global_vars:
                 a = ','.join( local_vars-global_vars )
                 writer.write('var(%s)' %a)
 
+        if self._with_js or javascript:
+            if node.args.defaults:
+                for i, arg in enumerate(node.args.args):
+                    dindex = i - len(node.args.defaults)
+                    if dindex >= 0:
+                        default_value = self.visit( node.args.defaults[dindex] )
+                        writer.write("""JS("var %s = %s  || %s ")""" % (arg.id, arg.id, default_value))
 
 
-        if not self._with_js and (len(node.args.defaults) or len(node.args.args) or node.args.vararg or node.args.kwarg):
+        elif self._with_fastdef or fastdef:
+            for i, arg in enumerate(node.args.args):
+                dindex = i - len(node.args.defaults)
+                if dindex >= 0 and node.args.defaults:
+                    default_value = self.visit( node.args.defaults[dindex] )
+                    writer.write("""JS("var %s = kwargs[ '%s' ]  || %s ")""" % (arg.id, arg.id, default_value))
+                else:
+                    writer.write("""JS("var %s = args[ %s ]")""" % (arg.id, i))
+
+        elif len(node.args.defaults) or len(node.args.args) or node.args.vararg or node.args.kwarg:
             # First check the arguments are well formed 
             # ie. that this function is not a callback of javascript code
             writer.write("""if (JS('args instanceof Array') and JS("{}.toString.call(kwargs) === '[object Object]'") and arguments.length == 2):""")
@@ -1150,11 +1264,42 @@ class PythonToPythonJS(NodeVisitor):
         else:
             log('(function has no arguments)')
 
-        self._return_type = None
-        map(self.visit, node.body)
+        ################# function body #################
+        if self._cached_property:
+            writer.write('if self["__dict__"]["%s"]: return self["__dict__"]["%s"]' %(self._cached_property, self._cached_property))
 
-        if self._return_type:
+
+        self._return_type = None # tries to catch a return type in visit_Return
+
+        map(self.visit, node.body)  ## write function body
+
+        if self._return_type:       ## check if a return type was caught
+            if return_type:
+                assert return_type == self._return_type
+            else:
+                return_type = self._return_type
             self._function_return_types[ node.name ] = self._return_type
+        self._return_type = None
+
+        if setter and 'set' in self._injector:  ## inject extra code
+            value_name = node.args.args[1].id
+            inject = [
+                'if self.__dict__.property_callbacks["%s"]:' %prop_name,
+                'self.__dict__.property_callbacks["%s"](["%s", %s, self], JSObject())' %(prop_name, prop_name, value_name)
+            ]
+            writer.write( ' '.join(inject) )
+
+        elif self._injector and node.original_name == '__init__':
+            if 'set' in self._injector:
+                writer.write( 'self["__dict__"]["property_callbacks"] = JSObject()' )
+            if 'init' in self._injector:
+                writer.write('if self.__class__.init_callbacks.length:')
+                writer.push()
+                writer.write('for callback in self.__class__.init_callbacks:')
+                writer.push()
+                writer.write('callback( [self], JSObject() )')
+                writer.pull()
+                writer.pull()
 
         writer.pull()
         ## note, in javascript function.name is a non-standard readonly attribute,
@@ -1165,6 +1310,21 @@ class PythonToPythonJS(NodeVisitor):
         defaults = ['%s:%s'%(self.visit(x[0]), self.visit(x[1])) for x in zip(node.args.args[-len(node.args.defaults):], node.args.defaults) ]
         writer.write( '%s.kwargs_signature = {%s}' %(node.name, ','.join(defaults)) )
 
+        #types = ['%s:%s'%(self.visit(x[0]), '"%s"'%type(self.visit(x[1])).__name__ ) for x in zip(node.args.args[-len(node.args.defaults):], node.args.defaults) ]
+        types = []
+        for x in zip(node.args.args[-len(node.args.defaults):], node.args.defaults):
+            key = x[0]
+            value = x[1]
+            if isinstance(value, ast.Name):
+                value = value.id
+            else:
+                value = type(value).__name__.lower()
+            types.append( '%s : "%s"' %(self.visit(key), value) )
+
+        writer.write( '%s.types_signature = {%s}' %(node.name, ','.join(types)) )
+        if return_type:
+            writer.write('%s.return_type = "%s"'%(node.name, return_type))
+
         if self._with_js and with_js_decorators:
             for dec in with_js_decorators:
                 if '.prototype.' in dec:
@@ -1182,19 +1342,27 @@ class PythonToPythonJS(NodeVisitor):
         #elif self._with_js:  ## this is just an optimization so we can avoid making wrappers at runtime
         #    writer.write('%s.pythonscript_function=true'%node.name)
 
-        if not self._with_js:
+        if not self._with_js and not javascript:
             writer.write('%s.pythonscript_function=True'%node.name)
 
         # apply decorators
         for decorator in decorators:
             assert not self._with_js
-            writer.write('%s = get_attribute(%s,"__call__")( [%s], {} )' % (node.name, self.visit(decorator), node.name))
+            writer.write('%s = get_attribute(%s,"__call__")( [%s], JSObject() )' % (node.name, self.visit(decorator), node.name))
+
+    #################### loops ###################
+    ## the old-style for loop that puts a while loop inside a try/except and catches StopIteration,
+    ## has a problem because at runtime if there is an error inside the loop, it will not show up in a strack trace,
+    ## the error is slient.  FAST_FOR is safer and faster, although it is not strictly Python because in standard
+    ## Python a list is allowed to grow or string while looping over it.  FAST_FOR only deals with a fixed size thing to loop over.
+    FAST_FOR = True
 
     def visit_Continue(self, node):
         if self._with_js:
             writer.write('continue')
         else:
-            writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % self._for_iterator_target)
+            if not self.FAST_FOR:
+                writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % self._for_iterator_target)
             writer.write('continue')
         return ''
 
@@ -1212,20 +1380,48 @@ class PythonToPythonJS(NodeVisitor):
 
             self._for_iterator_target = node.target.id  ## this could break with nested for loops
             writer.write('var(__iterator__, %s)' % node.target.id)
-            writer.write('__iterator__ = get_attribute(get_attribute(%s, "__iter__"), "__call__")(JSArray(), JSObject())' % self.visit(node.iter))
-            writer.write('try:')
-            writer.push()
-            writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % node.target.id)
-            writer.write('while True:')
-            writer.push()
-            map(self.visit, node.body)
-            writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % node.target.id)
-            writer.pull()
-            writer.pull()
-            writer.write('except StopIteration:')
-            writer.push()
-            writer.write('pass')
-            writer.pull()
+    
+            is_range = False
+            if self.FAST_FOR and isinstance(node.iter, ast.Call) and isinstance(node.iter.func, Name) and node.iter.func.id == 'range':
+                is_range = True
+            else:
+                writer.write('__iterator__ = get_attribute(get_attribute(%s, "__iter__"), "__call__")(JSArray(), JSObject())' % self.visit(node.iter))
+
+            if self.FAST_FOR:
+                if is_range:
+                    iter_name = node.target.id
+                    range_num = self.visit( node.iter.args[0] )
+                    writer.write('var(%s)' %iter_name)
+                    writer.write('%s = 0' %iter_name)
+                    writer.write('while %s < %s:' %(iter_name, range_num))
+                    writer.push()
+                    map(self.visit, node.body)
+                    writer.write('%s += 1' %iter_name )
+                    writer.pull()
+                else:
+                    writer.write('var(__next__)')
+                    writer.write('__next__ = get_attribute(__iterator__, "next_fast")')
+                    writer.write('while __iterator__.__dict__.index < __iterator__.__dict__.length:')
+                    writer.push()
+                    writer.write('%s = __next__()' % node.target.id)
+                    map(self.visit, node.body)
+                    writer.pull()
+
+            else:
+                writer.write('try:')
+                writer.push()
+                writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % node.target.id)
+                writer.write('while True:')
+                writer.push()
+                map(self.visit, node.body)
+                writer.write('%s = get_attribute(__iterator__, "next")(JSArray(), JSObject())' % node.target.id)
+                writer.pull()
+                writer.pull()
+                writer.write('except StopIteration:')
+                writer.push()
+                writer.write('pass')
+                writer.pull()
+    
             return ''
 
     def visit_While(self, node):
@@ -1250,14 +1446,21 @@ class PythonToPythonJS(NodeVisitor):
             writer.with_javascript = True
             self._with_js = True
 
+        elif isinstance( node.context_expr, Name ) and node.context_expr.id == 'fastdef':
+            self._with_fastdef = True
+            map(self.visit, node.body)
+            self._with_fastdef = False
+
+
         elif isinstance( node.context_expr, Name ) and node.optional_vars and isinstance(node.optional_vars, Name) and node.optional_vars.id == 'jsobject':
-            instance_name = node.context_expr.id
-            for n in node.body:
-                if isinstance(n, ast.Expr) and isinstance(n.value, Name):
-                    attr_name = n.value.id
-                    writer.write('%s.%s = get_attribute(%s, "%s")'%(instance_name, attr_name, instance_name, attr_name))
-                else:
-                    raise SyntaxError('invalid statement inside of "with x as jsobject:" block')
+            #instance_name = node.context_expr.id
+            #for n in node.body:
+            #    if isinstance(n, ast.Expr) and isinstance(n.value, Name):
+            #        attr_name = n.value.id
+            #        writer.write('%s.%s = get_attribute(%s, "%s")'%(instance_name, attr_name, instance_name, attr_name))
+            #    else:
+            #        raise SyntaxError('invalid statement inside of "with x as jsobject:" block')
+            raise SyntaxError('"with x as jsobject:" is DEPRECATED - methods on instances are now callable by default from JavaScript')
 
         else:
             raise SyntaxError('improper use of "with" statement')
