@@ -1,8 +1,8 @@
 #!/usr/bin/env python
 import sys
 
-from python_to_pythonjs import main as python_to_pythonjs
-from pythonjs import main as pythonjs
+from pythonjs.python_to_pythonjs import main as python_to_pythonjs
+from pythonjs.pythonjs import main as pythonjs
 
 
 def main(script):
